"""
Fetch metadata about streams, including stream types.

"""
import datetime
from io import StringIO
import json
from typing import Callable, Iterable, Iterator, List, Optional, Type, Union

import pandas as pd

from .client import GraphClient, StreamClient, global_graph_client
from .common import ItemBase, ItemSet
from .patient import Device, Patient, get_patient
from .stream import get_stream_availability, get_stream_data
from runeq.errors import RuneError


class Dimension(ItemBase):
    """
    A dimension of a stream type. This is akin to a column in a table,
    where each value in the timeseries is a row.

    """
    def __init__(
        self,
        id: str,
        data_type: str,
        quantity_name: str,
        unit_name: str,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: Dimension ID
            data_type: Data type (e.g. sfloat, timestamp, etc...)
            quantity_name: Name of the quantity measured
            unit_name: Name of the unit measurement

        """
        self.data_type = data_type
        self.quantity_name = quantity_name
        self.unit_name = unit_name

        super().__init__(
            id=id,
            data_type=data_type,
            quantity_name=quantity_name,
            unit_name=unit_name,
            **attributes,
        )


class StreamType(ItemBase):
    """
    A stream type to categorize streams.

    It represents the physical quantity being measured (voltage, acceleration,
    etc), including the unit of measurement. It also describes the shape of
    the stream’s data, as one or more dimensions.

    """

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        dimensions: List[Dimension],
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: StreamType ID
            name: Human-readable name of the stream type
            description: Human-readable description of the stream type
            dimensions: List of Dimensions in this stream type

        """
        self.name = name
        self.description = description
        self.dimensions = dimensions

        super().__init__(
            id=id,
            name=name,
            description=description,
            dimensions=dimensions,
            **attributes,
        )

    def to_dict(self) -> dict:
        """
        Dictionary representation of the Stream Type attributes.

        """
        attrs = self._attributes.copy()
        attrs["dimensions"] = [dim.to_dict() for dim in self.dimensions]
        return attrs


class StreamTypeSet(ItemSet):
    """
    A collection of StreamTypes.

    """

    def __init__(self, items: Iterable[StreamType] = ()):
        """
        Initialize with StreamTypes.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return StreamType


def get_all_stream_types(
    client: Optional[GraphClient] = None
) -> StreamTypeSet:
    """
    Get all stream types.

    Args:
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    query = '''
        query getStreamTypes {
            streamTypeList {
                streamTypes {
                    id
                    name
                    description
                    shape {
                        dimensions {
                            id: identifier
                            data_type: dataType
                            quantity_name: quantityName
                            quantity_abbrev: quantityAbbrev
                            unit_name: unitName
                            unit_abbrev: unitAbbrev
                        }
                    }
                }
            }
        }
    '''

    stream_type_set = StreamTypeSet()
    result = client.execute(statement=query)

    stream_type_list = result.get("streamTypeList", {})

    for stream_attrs in stream_type_list.get("streamTypes", []):
        stream_type = _parse_stream_type(stream_attrs)
        stream_type_set.add(stream_type)

    return stream_type_set


def _parse_stream_type(stream_type_attrs: dict) -> StreamType:
    """
    Parse stream type graphql response body.

    Args:
        stream_type_attrs: Attribute dictionary from a graph ql response
            representing a stream type.


    """
    # Create a dimension set from the stream's dimensions
    dimensions = []
    for dimension_attrs in stream_type_attrs["shape"].get("dimensions", []):
        dimensions.append(
            Dimension(**dimension_attrs)
        )

    del stream_type_attrs["shape"]

    return StreamType(
        dimensions=dimensions,
        **stream_type_attrs
    )


class StreamMetadata(ItemBase):
    """
    Metadata for a stream (i.e. timeseries data). This class also has methods
    that fetch data for the stream.

    """

    def __init__(
        self,
        id: str,
        created_at: float,
        algorithm: str,
        device_id: str,
        patient_id: str,
        stream_type: StreamType,
        min_time: float,
        max_time: float,
        parameters: dict,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: Stream ID
            created_at: When the stream was created, as a Unix timestamp, as
                a Unix timestamp (in seconds).
            algorithm: A versioned label that describes the process that was
                used to derive this timeseries.
            device_id: Device ID
            patient_id: Patient ID
            stream_type: Stream type, which categorizes the stream data.
            min_time: The earliest timestamp in the stream, as a Unix timestamp
                (in seconds).
            max_time: The latest timestamp in the stream, as a Unix timestamp
                (in seconds).
            parameters: Key/value pairs that label the stream.

        """
        self.created_at = created_at
        self.algorithm = algorithm
        self.device_id = device_id
        self.patient_id = patient_id
        self.stream_type = stream_type
        self.min_time = min_time
        self.max_time = max_time
        self.parameters = parameters

        super().__init__(
            id=id,
            created_at=created_at,
            algorithm=algorithm,
            device_id=device_id,
            patient_id=patient_id,
            stream_type=stream_type,
            min_time=min_time,
            max_time=max_time,
            parameters=parameters,
            **attributes
        )

    def to_dict(self) -> dict:
        """
        Dictionary representation of the StreamMetadata attributes.

        """
        attrs = self._attributes.copy()
        attrs["stream_type"] = self.stream_type.to_dict()
        return attrs

    def iter_stream_data(
        self,
        start_time: Optional[Union[float, datetime.date]] = None,
        start_time_ns: Optional[int] = None,
        end_time: Optional[Union[float, datetime.date]] = None,
        end_time_ns: Optional[int] = None,
        format: Optional[str] = "csv",
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        translate_enums: Optional[bool] = True,
        client: Optional[StreamClient] = None
    ) -> Iterator[Union[str, dict]]:
        """
        Iterate over CSV-formatted data for this stream.

        Args:
            start_time: Start time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
            start_time_ns: Start time for the query, provided as a unix
                timestamp (in nanoseconds).
            end_time: End time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
            end_time_ns: End time for the query, provided as a unix timestamp
                (in nanoseconds).
            format: Either "csv" (default) or "json". Determines the content
                type of the API response, as well as the type that
                is returned from this function.
            limit: Maximum number of timestamps to return, across *all pages*
                of the response. A limit of 0 (default) will fetch all
                available data.
            page_token: Token to fetch the subsequent page of results.
                The value is obtained from the "X-Rune-Next-Page-Token"
                response header field.
            timestamp: One of "unix", "unixns", or "iso", which determines
                how timestamps are formatted in the response
            timezone: Timezone offset, in seconds, used to calculate
                string-based timestamp formats such as datetime and iso.
                For example, PST (UTC-0800) is represented as -28800.
                If omitted, the timezone is UTC.
            translate_enums: If True, enum values are returned as their string
                representation. Otherwise, enums are returned as integer
                values.
            client: If specified, this client is used to fetch data from the
                API. Otherwise, the global
                :class:`~runeq.resources.client.StreamClient` is used.

        Returns:
            An iterator over paginated API responses. If format is "json", each
            response is a dict. If format is "csv", each response is a
            CSV-formatted string.

        """
        return get_stream_data(
            stream_id=self.id,
            start_time=start_time,
            start_time_ns=start_time_ns,
            end_time=end_time,
            end_time_ns=end_time_ns,
            format=format,
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            translate_enums=translate_enums,
            client=client,
        )

    def _add_metadata_to_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Returns a new dataframe, with metadata columns added.

        """
        return df.assign(
            stream_id=self.id,
            stream_type_id=self.stream_type.id,
            patient_id=self.patient_id,
            device_id=self.device_id,
            algorithm=self.algorithm,
            **self.parameters,
        )

    def get_stream_dataframe(
        self,
        start_time: Optional[Union[float, datetime.date]] = None,
        start_time_ns: Optional[int] = None,
        end_time: Optional[Union[float, datetime.date]] = None,
        end_time_ns: Optional[int] = None,
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        translate_enums: Optional[bool] = True,
        stream_client: Optional[StreamClient] = None
    ) -> pd.DataFrame:
        """
        Get stream data as an enriched Pandas dataframe. In addition to the raw
        stream data, the dataframe also includes columns with stream metadata.
        This allows for the concatenation of dataframes with data from
        multiple streams.

        Args:
            start_time: Start time for the query, provided as a unix timestamp
                    (in seconds) or a datetime.date.
            start_time_ns: Start time for the query, provided as a unix
                timestamp (in nanoseconds).
            end_time: End time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
            end_time_ns: End time for the query, provided as a unix timestamp
                (in nanoseconds).format: Optional enum "json" or "csv", which
                determines the content type of the response
            limit: Maximum number of timestamps to return, across *all pages*
                of the response. A limit of 0 (default) will fetch all
                available data.
            page_token: Token to fetch the subsequent page of results.
                The value is obtained from the 'X-Rune-Next-Page-Token'
                response header field.
            timestamp: Optional enum "unix", "unixns", or "iso", which
                determines how timestamps are formatted in the response
            timezone: Optional timezone offset, in seconds, used to calculate
                string-based timestamp formats such as datetime and iso.
                For example, PST (UTC-0800) is represented as -28800.
                If omitted, the timezone is UTC.
            translate_enums: If True, enum values are returned as their string
                representation. Otherwise, enums are returned as integer
                values.
            stream_client: If specified, this client is used to fetch data
                from the API. Otherwise, the global StreamClient is used.

        """
        params = {
            'stream_id': self.id,
            'start_time': start_time,
            'start_time_ns': start_time_ns,
            'end_time': end_time,
            'end_time_ns': end_time_ns,
            'format': 'csv',
            'limit': limit,
            'page_token': page_token,
            'timestamp': timestamp,
            'timezone': timezone,
            'translate_enums': translate_enums,
        }

        all_stream_dfs = []
        for resp in get_stream_data(client=stream_client, **params):
            all_stream_dfs.append(pd.read_csv(StringIO(resp), sep=","))

        stream_df = pd.concat(all_stream_dfs, axis=0, ignore_index=True)

        # Convert "dict" dimensions to native dicts (from strings)
        for dim in self.stream_type.dimensions:
            if dim.data_type == 'dict':
                stream_df[dim.id] = stream_df[dim.id].apply(json.loads)

        # Add metadata before returning the dataframe
        return self._add_metadata_to_dataframe(stream_df)

    def get_stream_availability_dataframe(
        self,
        start_time: Union[float, datetime.date],
        end_time: Union[float, datetime.date],
        resolution: int,
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = 'iso',
        timezone: Optional[int] = None,
        stream_client: Optional[StreamClient] = None
    ) -> pd.DataFrame:
        """
        Get stream availability as an enriched Pandas dataframe. The dataframe
        includes columns with metadata for this stream. This allows for the
        concatenation of dataframes with availability for multiple streams.

        Args:
            start_time: Start time for the query, provided as a unix timestamp
                    (in seconds) or a datetime.date.
            end_time: End time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
            resolution: Interval between returned timestamps, in seconds.
            limit: Maximum number of timestamps to return, across *all pages*
                of the response. A limit of 0 (default) will fetch all
                available data.
            page_token: Token to fetch the subsequent page of results.
                The value is obtained from the 'X-Rune-Next-Page-Token'
                response header field.
            timestamp: Optional enum "unix", "unixns", or "iso", which
                determines how timestamps are formatted in the response
            timezone: Optional timezone offset, in seconds, used to calculate
                string-based timestamp formats such as datetime and iso.
                For example, PST (UTC-0800) is represented as -28800.
                If omitted, the timezone is UTC.
            stream_client: If specified, this client is used to fetch data
                from the API. Otherwise, the global StreamClient is used.

        """
        responses = get_stream_availability(
            stream_ids=self.id,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            format='csv',
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            client=stream_client,
        )

        all_stream_dfs = []
        for resp in responses:
            all_stream_dfs.append(pd.read_csv(StringIO(resp), sep=","))

        stream_df = pd.concat(all_stream_dfs, axis=0, ignore_index=True)
        # Add metadata before returning the dataframe
        return self._add_metadata_to_dataframe(stream_df)


class StreamMetadataSet(ItemSet):
    """
    A collection of StreamMetadata.

    """

    def __init__(self, items: Iterable[StreamMetadata] = ()):
        """
        Initialize with StreamMetadatas

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return StreamMetadata

    def filter(
        self,
        stream_id: Optional[str] = None,
        patient_id: Optional[str] = None,
        device_id: Optional[str] = None,
        stream_type_id: Optional[str] = None,
        algorithm: Optional[str] = None,
        category: Optional[str] = None,
        measurement: Optional[str] = None,
        filter_function: Optional[Callable[[StreamMetadata], bool]] = None,
        **parameters
    ) -> 'StreamMetadataSet':
        """
        Filters streams for those that match ALL optional filter parameters.
        Returns a new StreamMetadataSet.

        Args:
            stream_id: Stream ID
            patient_id: Patient ID
            device_id: Device ID
            stream_type_id: Stream type ID
            algorithm: A versioned label that describes the process that was
                used to derive this timeseries.
            category: A broad categorization of the data type (e.g. neural,
                vitals, etc)
            measurement: A specific label for what is being measured
                (e.g. heart_rate, step_count, etc).
            filter_function: User-defined filter function which accepts a
                Stream as a single argument and returns a boolean indicating
                whether to keep that stream.

        """
        new_stream_set = StreamMetadataSet()

        for stream in self._items.values():
            if (
                (not stream_id or stream.id == stream_id) and
                (not patient_id or stream.patient_id == patient_id) and
                (not device_id or stream.device_id == device_id) and
                (
                    not stream_type_id or
                    stream.stream_type.id == stream_type_id
                ) and
                (not algorithm or stream.algorithm == algorithm) and
                (not category or stream.get("category") == category) and
                (
                    not measurement or
                    stream.get("measurement") == measurement
                ) and
                all(
                    stream.get(param_name) == param
                    for param_name, param in parameters.items()
                ) and
                (not filter_function or filter_function(stream))
            ):
                new_stream_set.add(stream)

        return new_stream_set

    def get_stream_dataframe(
        self,
        start_time: Optional[Union[float, datetime.date]] = None,
        start_time_ns: Optional[int] = None,
        end_time: Optional[Union[float, datetime.date]] = None,
        end_time_ns: Optional[int] = None,
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = "iso",
        timezone: Optional[int] = None,
        translate_enums: Optional[bool] = True,
        stream_client: Optional[StreamClient] = None
    ) -> pd.DataFrame:
        """
        Get raw data for all streams in the collection, as an enriched Pandas
        dataframe. The dataframe includes columns with metadata for each
        stream.

        Args:
            start_time: Start time for the query, provided as a unix timestamp
                    (in seconds) or a datetime.date.
            start_time_ns: Start time for the query, provided as a unix
                timestamp (in nanoseconds).
            end_time: End time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
            end_time_ns: End time for the query, provided as a unix timestamp
                (in nanoseconds).format: Optional enum "json" or "csv", which
                determines the content type of the response
            limit: Maximum number of timestamps to return, across *all pages*
                of the response. A limit of 0 (default) will fetch all
                available data.
            page_token: Token to fetch the subsequent page of results.
                The value is obtained from the 'X-Rune-Next-Page-Token'
                response header field.
            timestamp: Optional enum "unix", "unixns", or "iso", which
                determines how timestamps are formatted in the response
            timezone: Optional timezone offset, in seconds, used to calculate
                string-based timestamp formats such as datetime and iso.
                For example, PST (UTC-0800) is represented as -28800.
                If omitted, the timezone is UTC.
            translate_enums: If True, enum values are returned as their string
                representation. Otherwise, enums are returned as integer
                values.
            stream_client: If specified, this client is used to fetch data
                from the API. Otherwise, the global StreamClient is used.

        """
        all_stream_dfs = []
        for stream_meta in self._items.values():
            stream_df = stream_meta.get_stream_dataframe(
                start_time=start_time,
                start_time_ns=start_time_ns,
                end_time=end_time,
                end_time_ns=end_time_ns,
                limit=limit,
                page_token=page_token,
                timestamp=timestamp,
                timezone=timezone,
                translate_enums=translate_enums,
                stream_client=stream_client,
            )
            all_stream_dfs.append(stream_df)

        return pd.concat(all_stream_dfs, axis=0, ignore_index=True)

    def get_batch_availability_dataframe(
        self,
        start_time: Union[float, datetime.date],
        end_time: Union[float, datetime.date],
        resolution: int,
        batch_operation: Optional[str] = None,
        limit: Optional[int] = None,
        page_token: Optional[str] = None,
        timestamp: Optional[str] = 'iso',
        timezone: Optional[int] = None,
        stream_client: Optional[StreamClient] = None
    ) -> pd.DataFrame:
        """
        Get stream availability data as a Pandas dataframe. Depending on the
        specified **batch_operation**, this fetches the availability of **all**
        or **any** of the streams in the collection.

        If you want to gather the availability of each *individual* stream in a
        dataframe, see get_stream_availability_dataframe()

        Args:
            start_time: Start time for the query, provided as a unix timestamp
                    (in seconds) or a datetime.date.
            end_time: End time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
            resolution: Interval between returned timestamps, in seconds.
            batch_operation: Either "any" or "all", which determines
                what type of batch calculation will determine availability for
                the batch of streams. Availability values will equal 1 when
                data is available for "all" or "any" of the requested streams
                in the given interval.
            limit: Maximum number of timestamps to return, across *all pages*
                of the response. A limit of 0 (default) will fetch all
                available data.
            page_token: Token to fetch the subsequent page of results.
                The value is obtained from the 'X-Rune-Next-Page-Token'
                response header field.
            timestamp: Optional enum "unix", "unixns", or "iso", which
                determines how timestamps are formatted in the response
            timezone: Optional timezone offset, in seconds, used to calculate
                string-based timestamp formats such as datetime and iso.
                For example, PST (UTC-0800) is represented as -28800.
                If omitted, the timezone is UTC.
            stream_client: If specified, this client is used to fetch data
                from the API. Otherwise, the global StreamClient is used.

        """
        responses = get_stream_availability(
            stream_ids=self.ids(),
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            batch_operation=batch_operation,
            format='csv',
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            client=stream_client,
        )

        all_stream_dfs = []
        for resp in responses:
            all_stream_dfs.append(pd.read_csv(StringIO(resp), sep=","))

        stream_df = pd.concat(all_stream_dfs, axis=0, ignore_index=True)

        return stream_df


def get_stream_metadata(
    stream_ids: Union[str, Iterable[str]],
    client: Optional[GraphClient] = None
) -> Union[StreamMetadata, StreamMetadataSet]:
    """
    Get stream metadata for the specified stream_id(s).

    Args:
        stream_ids: ID of the stream or list of IDs
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    Returns:
        StreamMetadata: if a single stream ID is specified
        StreamMetadataSet: if multiple stream IDs are specified

    Raises:
        RuneError: if any of the stream IDs are not found

    """
    client = client or global_graph_client()
    query = '''
        query getStreamListByIds($stream_ids: [String]) {
            streamListByIds(streamIds: $stream_ids) {
              pageInfo {
                  endCursor
              }
              streams {
                id
                created_at: createdAt
                algorithm
                device_id: deviceId
                patient_id: patientId
                streamType {
                  id
                  name
                  description
                  shape {
                        dimensions {
                            id: identifier
                            data_type: dataType
                            quantity_name: quantityName
                            quantity_abbrev: quantityAbbrev
                            unit_name: unitName
                            unit_abbrev: unitAbbrev
                        }
                    }
                }
                parameters {
                    key
                    value
                }
                min_time: minTime
                max_time: maxTime
              }
            }
        }
    '''
    stream_set = StreamMetadataSet()

    if type(stream_ids) is str:
        stream_ids = [stream_ids]
    else:
        stream_ids = list(stream_ids)

    # Query stream list in batches of <= 100 streams, since graph api cannot
    # query for more than 100 streams at once.
    stream_list_results = []
    for start in range(0, len(stream_ids), 100):
        result = client.execute(
            statement=query,
            stream_ids=stream_ids[start:start + 100],
        )
        stream_list_results.append(result)

    seen_stream_ids = set(stream_ids)
    for result in stream_list_results:
        stream_list = result.get("streamListByIds", {})
        for stream_attrs in stream_list.get("streams", []):
            stream_type = _parse_stream_type(stream_attrs["streamType"])

            del stream_attrs["streamType"]
            norm_dev_id = Device.normalize_id(stream_attrs["device_id"])
            stream_attrs["device_id"] = norm_dev_id

            # Add query parameters to stream attributes
            params = {}
            if stream_attrs.get("parameters"):
                for param in stream_attrs['parameters']:
                    stream_attrs[param["key"]] = param["value"]
                    params[param["key"]] = param["value"]
                del stream_attrs['parameters']

            stream = StreamMetadata(
                stream_type=stream_type,
                parameters=params,
                **stream_attrs
            )
            stream_set.add(stream)

            try:
                seen_stream_ids.remove(stream.id)
            except KeyError:
                pass

    if len(seen_stream_ids) > 0:
        raise RuneError(
            f"1+ stream ID(s) not found: {','.join(seen_stream_ids)}"
        )

    if len(stream_set) == 1:
        return list(stream_set)[0]

    return stream_set


def get_patient_stream_metadata(
    patient_id: str,
    device_id: Optional[str] = None,
    stream_type_id: Optional[str] = None,
    algorithm: Optional[str] = None,
    category: Optional[str] = None,
    measurement: Optional[str] = None,
    client: Optional[GraphClient] = None,
    **parameters
) -> StreamMetadataSet:
    """
    Get stream metadata for a patient's streams, matching ALL filter
    parameters. Only the patient ID is required.

    Args:
        patient_id: Patient ID
        device_id: Device ID
        stream_type_id: Stream type ID
        algorithm: A versioned label that describes the process that was
            used to derive this timeseries.
        category: A broad categorization of the data type (e.g. neural,
            vitals, etc)
        measurement: A specific label for what is being measured
            (e.g. heart_rate, step_count, etc).
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.
        **parameters: Key/value pairs that label the stream.

    """
    if not patient_id:
        raise ValueError("must provide a patient_id")

    # Checks whether patient is accessible, otherwise raises NotFoundError.
    get_patient(patient_id=patient_id, client=client)

    client = client or global_graph_client()
    query = '''
        query getStreamList($cursor: Cursor, $filters: StreamQueryFilters) {
            streamList(filters: $filters, cursor: $cursor) {
                pageInfo {
                    endCursor
                }
                streams {
                    id
                    created_at: createdAt
                    algorithm
                    device_id: deviceId
                    patient_id: patientId
                    streamType {
                        id
                        name
                        description
                        shape {
                            dimensions {
                                id: identifier
                                data_type: dataType
                                quantity_name: quantityName
                                quantity_abbrev: quantityAbbrev
                                unit_name: unitName
                                unit_abbrev: unitAbbrev
                            }
                        }
                    }
                    parameters {
                        key
                        value
                    }
                    min_time: minTime
                    max_time: maxTime
                }
            }
        }
    '''
    patient_id = Patient.normalize_id(patient_id)

    if device_id:
        device_id = Device.denormalize_id(patient_id, device_id)

    # Add category to params and format params list for filter query
    if category:
        parameters["category"] = category

    # Add measurement to params and format params list for filter query
    if measurement:
        parameters["measurement"] = measurement

    params = []
    for key, val in parameters.items():
        params.append(
            {
                "key": key,
                "value": val,
            }
        )

    filters = {
        "patientId": patient_id,
        "deviceId": device_id,
        "streamTypeId": stream_type_id,
        "algorithm": algorithm,
        "parameters": params
    }

    next_cursor = None
    stream_set = StreamMetadataSet()

    # Use cursor to page through all filtered streams
    while True:
        result = client.execute(
            statement=query,
            filters=filters,
            cursor=next_cursor
        )

        stream_list = result.get("streamList", {})
        for stream_attrs in stream_list.get("streams", []):
            stream_type = _parse_stream_type(stream_attrs["streamType"])

            del stream_attrs["streamType"]
            norm_dev_id = Device.normalize_id(stream_attrs["device_id"])
            stream_attrs["device_id"] = norm_dev_id

            # Add query parameters to stream attributes
            params = {}
            if stream_attrs.get("parameters"):
                for param in stream_attrs['parameters']:
                    stream_attrs[param["key"]] = param["value"]
                    params[param["key"]] = param["value"]
                del stream_attrs['parameters']

            stream = StreamMetadata(
                stream_type=stream_type,
                parameters=params,
                **stream_attrs
            )
            stream_set.add(stream)

        # next_cursor is None when there are no more streams
        next_cursor = stream_list.get("pageInfo", {}).get("endCursor")
        if not next_cursor:
            break

    return stream_set


def get_stream_dataframe(
    stream_ids: Union[str, Iterable[str]],
    start_time: Optional[Union[float, datetime.date]] = None,
    start_time_ns: Optional[int] = None,
    end_time: Optional[Union[float, datetime.date]] = None,
    end_time_ns: Optional[int] = None,
    limit: Optional[int] = None,
    page_token: Optional[str] = None,
    timestamp: Optional[str] = "iso",
    timezone: Optional[int] = None,
    translate_enums: Optional[bool] = True,
    stream_client: Optional[StreamClient] = None,
    graph_client: Optional[GraphClient] = None
) -> pd.DataFrame:
    """
    Get stream(s) as enriched dataframe with stream data and metadata.

    Args:
        stream_ids: 1 or multiple stream IDs
        start_time: Start time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
        start_time_ns: Start time for the query, provided as a unix timestamp
            (in nanoseconds).
        end_time: End time for the query, provided as a unix timestamp
            (in seconds) or a datetime.date.
        end_time_ns: End time for the query, provided as a unix timestamp
            (in nanoseconds).format: Optional enum "json" or "csv", which
            determines the content type of the response
        limit: Maximum number of timestamps to return, across *all pages*
            of the response. A limit of 0 (default) will fetch all
            available data.
        page_token: Token to fetch the subsequent page of results.
            The value is obtained from the 'X-Rune-Next-Page-Token'
            response header field.
        timestamp: Optional enum "unix", "unixns", or "iso", which determines
            how timestamps are formatted in the response
        timezone: Optional timezone offset, in seconds, used to calculate
            string-based timestamp formats such as datetime and iso.
            For example, PST (UTC-0800) is represented as -28800.
            If omitted, the timezone is UTC.
        translate_enums: If True, enum values are returned as their string
            representation. Otherwise, enums are returned as integer values.
        stream_client: If specified, this client is used to fetch data from the
            API. Otherwise, the global StreamClient is used.
        graph_client: If specified, this client is used to fetch metadata from
            the API. Otherwise, the global GraphClient is used.

    Raises:
        RuneError: if any of the stream IDs is not found.

    """
    stream_meta_set = get_stream_metadata(
        stream_ids=stream_ids,
        client=graph_client
    )

    return stream_meta_set.get_stream_dataframe(
        start_time=start_time,
        start_time_ns=start_time_ns,
        end_time=end_time,
        end_time_ns=end_time_ns,
        limit=limit,
        page_token=page_token,
        timestamp=timestamp,
        timezone=timezone,
        translate_enums=translate_enums,
        stream_client=stream_client,
    )


def get_stream_availability_dataframe(
    stream_ids: Union[str, Iterable[str]],
    start_time: Union[float, datetime.date],
    end_time: Union[float, datetime.date],
    resolution: int,
    batch_operation: Optional[str] = None,
    limit: Optional[int] = None,
    page_token: Optional[str] = None,
    timestamp: Optional[str] = 'iso',
    timezone: Optional[int] = None,
    stream_client: Optional[StreamClient] = None,
    graph_client: Optional[GraphClient] = None
) -> pd.DataFrame:
    """
    Get stream availability data as a dataframe. If a single stream_id is
    passed in, the dataframe will be enriched with the stream's metadata.
    Otherwise it will contain just the availability data.

    Note that this is different from the get_stream_availability_dataframe
    method on the :class:`~runeq.resources.stream_metadata.StreamMetadataSet`.
    This dataframe contains the availability of each *individual* stream.

    Args:
        stream_ids: 1 or multiple stream IDs
        start_time: Start time for the query, provided as a unix timestamp
                (in seconds) or a datetime.date.
        end_time: End time for the query, provided as a unix timestamp
            (in seconds) or a datetime.date.
        resolution: Interval between returned timestamps, in seconds.
        batch_operation: Either "any" or "all", which determines
            what type of batch calculation will determine availability for the
            batch of streams. Availability values will equal 1 when
            data is available for "all" or "any" of the requested streams in
            the given interval.
        limit: Maximum number of timestamps to return, across *all pages*
            of the response. A limit of 0 (default) will fetch all
            available data.
        page_token: Token to fetch the subsequent page of results.
            The value is obtained from the 'X-Rune-Next-Page-Token'
            response header field.
        timestamp: Optional enum "unix", "unixns", or "iso", which determines
            how timestamps are formatted in the response
        timezone: Optional timezone offset, in seconds, used to calculate
            string-based timestamp formats such as datetime and iso.
            For example, PST (UTC-0800) is represented as -28800.
            If omitted, the timezone is UTC.
        stream_client: If specified, this client is used to fetch data from the
            API. Otherwise, the global StreamClient is used.
        graph_client: If specified, this client is used to fetch metadata from
            the API. Otherwise, the global GraphClient is used.

    """
    # Standardize iterable stream_ids to a list (so that we can use len)
    if type(stream_ids) is not str:
        stream_ids = list(stream_ids)

    # If there are multiple stream ids, there is no need to get metadata
    # since the dataframe response will be simplified
    if type(stream_ids) is list and len(stream_ids) > 1:
        responses = get_stream_availability(
            stream_ids=stream_ids,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            batch_operation=batch_operation,
            format='csv',
            limit=limit,
            page_token=page_token,
            timestamp=timestamp,
            timezone=timezone,
            client=stream_client,
        )

        all_stream_dfs = []
        for resp in responses:
            all_stream_dfs.append(pd.read_csv(StringIO(resp), sep=","))

        stream_df = pd.concat(all_stream_dfs, axis=0, ignore_index=True)

        return stream_df

    # Since there is only one stream id, we can enrich the dataframe with
    # metadata
    stream_meta = get_stream_metadata(
        stream_ids=stream_ids,
        client=graph_client
    )

    return stream_meta.get_stream_availability_dataframe(
        start_time=start_time,
        end_time=end_time,
        resolution=resolution,
        limit=limit,
        page_token=page_token,
        timestamp=timestamp,
        timezone=timezone,
        stream_client=stream_client,
    )



from unittest import mock, TestCase
from typing import List, Dict

import numpy as np

from runeq import Config, stream, errors


def mock_get_json_response(
        bodies: List[dict],
        calls: List,
        status_code=200,
        headers: List[Dict[str, str]] = None
):
    """Return a function that can be used to mock .get_json_response()

    Args:
        bodies: list of JSON bodies to return from response.json()
        calls: list. Every time the function is called, args and kwargs will
            be appended to this list.
        status_code: status code to return for each response
        headers: the response header to apply
    """
    headers = headers or [{}] * len(bodies)

    num = 0

    def _func(*args, **kwargs):
        nonlocal num
        # add inputs to the list of calls that was provided
        calls.append((args, kwargs))

        resp = mock.MagicMock()
        resp.headers = headers[num]
        resp.status_code = status_code
        resp.ok = (status_code < 400)
        resp.json.return_value = bodies[num]
        num += 1
        return resp

    return _func


def mock_get_csv_response(
        bodies: List[str],
        calls: List,
        status_code=200,
        headers: List[Dict[str, str]] = None
):
    """Return a function that can be used to mock .get_csv_response()

    Args:
        bodies: list of response bodies to return from response.text
        calls: list. Every time the function is called, args and kwargs will
            be appended to this list.
        status_code: status code to return for each response
        headers: the response header to apply
    """
    headers = headers or [{}] * len(bodies)

    num = 0

    def _func(*args, **kwargs):
        nonlocal num
        # keep track of the kwargs that were used to call this
        calls.append((args, kwargs))

        resp = mock.MagicMock()
        resp.headers = headers[num]
        resp.status_code = status_code
        resp.ok = (status_code < 400)
        resp.text = bodies[num]
        num += 1
        return resp

    return _func


class TestStreamV1Client(TestCase):
    """
    Test stream.V1Client and the associated accessors.

    """

    def setUp(self) -> None:
        """
        Initialize a client, set up basic mocking.

        """
        self.cfg = Config(
            client_key_id='abc',
            client_access_key='abc123',
        )
        self.client = stream.V1Client(self.cfg)
        self.use_np_orig = stream.v1.USE_NUMPY
        stream.v1.USE_NUMPY = False

    def tearDown(self) -> None:
        """
        Tear down monkey-patching.

        """
        stream.v1.USE_NUMPY = self.use_np_orig

    @mock.patch('runeq.stream.v1.requests')
    def test_get_json_response(self, requests):
        """
        Test the signature of JSON requests.

        """
        for test_num, case in enumerate((
                (self.client.Accel, '/v1/accel.json'),
                (self.client.BandPower, '/v1/band_power.json'),
                (self.client.Event, '/v1/event.json'),
                (self.client.HeartRate, '/v1/heartrate.json'),
                (self.client.LFP, '/v1/lfp.json'),
                (
                        self.client.ProbabilitySymptom,
                        '/v1/probability_symptom.json'
                ),
                (self.client.Rotation, '/v1/rotation.json'),
                (self.client.Span, '/v1/span.json'),
                (self.client.State, '/v1/state.json'),
        )):
            resource_creator, endpoint = case
            resource = resource_creator(leslie='knope')
            resource.get_json_response(ron='swanson', test_num=test_num)
            requests.get.assert_has_calls([
                mock.call(
                    self.cfg.stream_url + endpoint,
                    headers=self.cfg.auth_headers,
                    params={
                        'leslie': 'knope',
                        'ron': 'swanson',
                        'test_num': test_num,
                    }
                ),
            ])

    @mock.patch('runeq.stream.v1.requests')
    def test_get_csv_response(self, requests):
        """
        Test the signature of CSV requests.

        """
        for test_num, case in enumerate((
                (self.client.Accel, '/v1/accel.csv'),
                (self.client.BandPower, '/v1/band_power.csv'),
                (self.client.HeartRate, '/v1/heartrate.csv'),
                (self.client.LFP, '/v1/lfp.csv'),
                (
                        self.client.ProbabilitySymptom,
                        '/v1/probability_symptom.csv'
                ),
                (self.client.Rotation, '/v1/rotation.csv'),
                (self.client.State, '/v1/state.csv'),
        )):
            resource_creator, endpoint = case
            resource = resource_creator(leslie='knope')
            resource.get_csv_response(ron='swanson', test_num=test_num)
            requests.get.assert_has_calls([
                mock.call(
                    self.cfg.stream_url + endpoint,
                    stream=True,
                    headers=self.cfg.auth_headers,
                    params={
                        'leslie': 'knope',
                        'ron': 'swanson',
                        'test_num': test_num,
                    }
                ),
            ])

    def test_iter_json_data_with_token(self):
        """
        Test the iterator over JSON responses, paginating with the next page
        token header.

        """
        for test_num, resource_creator in enumerate((
                self.client.Accel,
                self.client.BandPower,
                self.client.Event,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.Span,
                self.client.State,
        )):
            resource = resource_creator()

            mock_responses = [
                {'success': True, 'result': [], 'next_page': 1},
                {'success': True, 'result': []}
            ]

            calls = []
            resource.get_json_response = mock_get_json_response(
                mock_responses,
                calls,
                200,
                [
                    {'X-Rune-Next-Page-Token': 'MTIzNDU2MDAwMA=='},
                    {},
                ]
            )

            iterator = resource.iter_json_data(test_num=test_num)
            self.assertEqual(len(list(iterator)), 2)

            # Check that all parameters were kept the same across calls,
            # except for "next_page_token"
            self.assertEqual(calls, [
                ((), {'test_num': test_num}),
                (
                    (),
                    {
                        'test_num': test_num,
                        'next_page_token': 'MTIzNDU2MDAwMA=='
                    }
                )
            ])

    def test_iter_json_data(self):
        """
        Test the iterator over JSON responses, following pagination with
        the page number.

        """
        results = [
            {'a': 1},
            {'b': 2}
        ]
        mock_responses = [
            {'success': True, 'result': results[0], 'next_page': 1},
            {'success': True, 'result': results[1]}
        ]

        for test_num, resource_creator in enumerate((
                self.client.Accel,
                self.client.BandPower,
                self.client.Event,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
        )):
            resource = resource_creator()

            #
            # Successful Requests
            #
            calls = []
            resource.get_json_response = mock_get_json_response(
                mock_responses,
                calls
            )

            # Check the results
            num_results = 0
            iterator = resource.iter_json_data(test_num=test_num)
            for i, actual in enumerate(iterator):
                self.assertEqual(results[i], actual)
                num_results += 1

            self.assertEqual(num_results, 2)

            # Check that all parameters were kept the same across calls,
            # except for "page" (which must be incremented)
            self.assertEqual(calls, [
                ((), {'test_num': test_num}),
                ((), {'test_num': test_num, 'page': 1})
            ])

            #
            # Request Error
            # Iterator should check the response status for each request
            #
            err_details = {
                "message": "i am an intentional error!",
                "type": "TestError",
            }
            resource.get_json_response = mock_get_json_response(
                [{'success': False, 'error': err_details}],
                [],
                status_code=404,
            )

            with self.assertRaises(errors.APIError) as e:
                next(resource.iter_json_data())

            err = e.exception
            self.assertEqual(err.status_code, 404)
            self.assertEqual(err.details, err_details)

    def test_iter_csv_data_with_token(self):
        """
        Test the iterator over CSV responses, which follows new pagination

        """
        mock_responses = [
            'good,better\nskiing,hiking\n',
            'good,better\ncupcakes,brownies\n',
            '',
        ]

        for test_num, resource_creator in enumerate((
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
        )):
            resource = resource_creator()

            #
            # Successful Requests
            #
            calls = []
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                calls,
                200,
                [
                    {'X-Rune-Next-Page-Token': 'MTIzNDU2MDAwMA=='},
                    {'X-Rune-Next-Page-Token': 'MTIzNDU2MDAwMA=='},
                    {},
                ],
            )

            # Check the results
            iterator = resource.iter_csv_text(test_num=test_num)
            self.assertEqual(len(list(iterator)), 2)

            # Check that all parameters were kept the same across calls,
            # except for "next_page_token" (which will normally be different
            # for each response)
            self.assertEqual(calls, [
                ((), {'test_num': test_num}),
                (
                    (),
                    {
                        'test_num': test_num,
                        'next_page_token': 'MTIzNDU2MDAwMA=='
                    }
                ),
                (
                    (),
                    {
                        'test_num': test_num,
                        'next_page_token': 'MTIzNDU2MDAwMA=='
                    }
                ),
            ])

    def test_iter_csv_data(self):
        """
        Test the iterator over CSV responses, which follows pagination
        """
        mock_responses = [
            'good,better\nskiing,hiking\n',
            'good,better\ncupcakes,brownies\n',
            '',
        ]

        for test_num, resource_creator in enumerate((
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
        )):
            resource = resource_creator()

            #
            # Successful Requests
            #
            calls = []
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                calls,
            )

            # Check the results
            num_results = 0
            iterator = resource.iter_csv_text(test_num=test_num)
            for i, actual in enumerate(iterator):
                self.assertEqual(mock_responses[i], actual)
                num_results += 1

            # although there are 3 responses, the last (empty) body should not
            # be returned by the iterator
            self.assertEqual(num_results, 2)

            # Check that all parameters were kept the same across calls,
            # except for "page" (which must be incremented)
            self.assertEqual(calls, [
                ((), {'test_num': test_num}),
                ((), {'test_num': test_num, 'page': 1}),
                ((), {'test_num': test_num, 'page': 2}),
            ])

            #
            # Request Error
            # Iterator should check the response status for each request
            #
            err_details = {
                "message": "i am an intentional error!",
                "type": "TestError",
            }
            # note: CSV endpoints return JSON on API errors
            resource.get_csv_response = mock_get_json_response(
                [{'success': False, 'error': err_details}],
                [],
                status_code=404,
            )

            with self.assertRaises(errors.APIError) as e:
                next(resource.iter_csv_text())

            err = e.exception
            self.assertEqual(err.status_code, 404)
            self.assertEqual(err.details, err_details)

    def test_iter_points(self):
        """
        Test iterating over data as points. Uses the CSV endpoint.

        """
        mock_responses = [
            'lower,higher,label\n1,2,ints\n3.5,6.7,floats\n',
            'lower,higher,label\n,8.9,missing data\n',
            '',
        ]
        expected = [
            {'lower': 1, 'higher': 2, 'label': 'ints'},
            {'lower': 3.5, 'higher': 6.7, 'label': 'floats'},
            {'lower': None, 'higher': 8.9, 'label': 'missing data'},
        ]

        for test_num, resource_creator in enumerate((
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
        )):
            resource = resource_creator()
            # replace get_csv_response on the resource
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                [],
            )

            for i, point in enumerate(resource.points()):
                self.assertDictEqual(expected[i], point)

            # replace get_csv_response again, to restart the mock responses
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                [],
            )

            for i, point in enumerate(resource):
                self.assertDictEqual(expected[i], point)
                # check dtype for "higher", which always has a numeric
                # value in the test data
                self.assertNotIsInstance(point['higher'], np.float64)

    def test_iter_points_numpy(self):
        """
        Test iterating over data as points, using Numpy to convert.

        """
        stream.v1.USE_NUMPY = True

        mock_responses = [
            'lower,higher,label\n1,2,ints\n3.5,6.7,floats\n',
            'lower,higher,label\n,8.9,missing data\n',
            '',
        ]
        expected = [
            {'lower': 1, 'higher': 2, 'label': 'ints'},
            {'lower': 3.5, 'higher': 6.7, 'label': 'floats'},
            {'lower': np.NaN, 'higher': 8.9, 'label': 'missing data'},
        ]

        for test_num, resource_creator in enumerate((
                self.client.Accel,
                self.client.BandPower,
                self.client.HeartRate,
                self.client.LFP,
                self.client.ProbabilitySymptom,
                self.client.Rotation,
                self.client.State,
        )):
            resource = resource_creator()
            # replace get_csv_response on the resource
            resource.get_csv_response = mock_get_csv_response(
                mock_responses,
                [],
            )

            for i, point in enumerate(resource.points()):
                self.assertDictEqual(expected[i], point)
                # check dtype for "higher", which always has a numeric
                # value in the test data
                self.assertIsInstance(point['higher'], np.float64)



"""
The Rune Labs Standard Development Kit (SDK) for querying data from Rune
APIs.

"""
from . import stream, resources
from .config import Config
from .resources.client import initialize


__all__ = [
    'Config',
    'initialize',
    'resources',
    'stream',
]



"""
Fetch metadata about patients, including their devices.

"""

from typing import Iterable, List, Optional, Type, Union

from .client import GraphClient, global_graph_client
from .common import ItemBase, ItemSet


class Device(ItemBase):
    """
    A patient device.

    A device is a sensor serving as the datasource, such as a neural implant,
    phone, or wearable. Each device belongs to a patient. A patient may have
    multiple devices: each one has a unique identifier ("device ID").

    """

    def __init__(
        self,
        id: str,
        patient_id: str,
        name: str,
        created_at: float,
        device_type_id: str,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: ID of the device
            patient_id: ID of the patient
            name: Human-readable name for the device
            created_at: When the device was created (unix timestamp)
            **attributes: Other attributes associated with the device

        """
        self.patient_id = patient_id
        self.name = name
        self.created_at = created_at
        self.device_type_id = device_type_id

        super().__init__(
            id=id,
            patient_id=patient_id,
            name=name,
            created_at=created_at,
            device_type_id=device_type_id,
            **attributes,
        )

    @staticmethod
    def normalize_id(device_id: str) -> str:
        """
        Strip resource prefix and suffix from a device ID (if they exist).

        Args:
            device_id: Device ID
        """
        id = device_id.split(",")[-1]

        if id.startswith("device-"):
            id = id[7:]

        return id

    @staticmethod
    def denormalize_id(patient_id: str, device_id: str) -> str:
        """
        Add resource prefix and suffix to a patient/device ID.

        This constructs the form of the ID that is used for requests to the
        GraphQL API.

        Args:
            patient_id: ID of the patient who owns the device
            device_id: Device ID

        """
        norm_patient_id = Patient.normalize_id(patient_id)
        norm_device_id = Device.normalize_id(device_id)

        return f'patient-{norm_patient_id},device-{norm_device_id}'

    def __repr__(self):
        """
        Override repr, to include the patient ID (as well as the device ID
        and name)

        """
        return (
            f'{self.__class__.__name__}(id="{self.id}", '
            f'name="{self.name}", patient_id="{self.patient_id}")'
        )


class DeviceSet(ItemSet):
    """
    A collection of Devices.

    """

    def __init__(self, items: Iterable[Device] = ()):
        """
        Initialize with Devices.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Device


class Patient(ItemBase):
    """
    A patient.

    A patient represents a person about whom data was measured. This may
    include a StrivePD user, a person with a neural implant, etc. Each patient
    has a unique identifier ("patient ID").

    """

    def __init__(
        self,
        id: str,
        name: str,
        created_at: float,
        devices: DeviceSet,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: ID of the patient
            name: Human-readable display name for the patient
            created_at: When the patient's record was created (unix timestamp)
            devices: Devices that belong to the patient
            **attributes: Other attributes associated with the patient
        """
        self.name = name
        self.created_at = created_at
        self.devices = devices

        super().__init__(
            id=id,
            name=name,
            created_at=created_at,
            devices=devices,
            **attributes,
        )

    @staticmethod
    def normalize_id(patient_id: str) -> str:
        """
        Strip resource prefix from a patient ID (if it exists).

        Args:
            patient_id: Patient ID

        """
        if patient_id.startswith("patient-"):
            patient_id = patient_id[8:]

        return patient_id

    @staticmethod
    def denormalize_id(patient_id: str) -> str:
        """
        Add resource prefix to a patient ID (if it doesn't exist).

        This constructs the form of the ID that is used for requests to the
        GraphQL API.

        Args:
            patient_id: Patient ID

        """
        if not patient_id.startswith("patient-"):
            patient_id = f'patient-{patient_id}'

        return patient_id

    def device(self, device_id: str) -> Device:
        """
        Return the patient's device with the specified device ID.

        Args:
            device_id: Device ID

        Raises:
            ValueError: if the patient does not have a device with the ID

        """
        device_id = Device.normalize_id(device_id)

        for device in self.devices:
            if device.id == device_id:
                return device

        raise ValueError("Device not found with id: %s" % device_id)

    def to_dict(self) -> dict:
        """
        Dictionary representation of the Patient attributes.

        """
        attrs = self._attributes.copy()
        attrs["devices"] = self.devices.to_list()
        return attrs


class PatientSet(ItemSet):
    """
    A collection of Patients.

    """

    def __init__(self, items: Iterable[Patient] = ()):
        """
        Initialize with Patients.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Patient

    @property
    def devices(self) -> DeviceSet:
        """
        Set of all devices that belong to the patients in this collection.

        """
        all_devices = DeviceSet()
        for patient in self._items.values():
            for device in patient.devices:
                all_devices.add(device)

        return all_devices


def get_patient(
    patient_id: str,
    client: Optional[GraphClient] = None
) -> Patient:
    """
    Get the patient with the specified patient ID.

    Args:
        patient_id: Patient ID
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    query = '''
        query getPatient($patient_id: ID!, $cursor: Cursor) {
            patient(id: $patient_id) {
                id
                name: codeName
                created_at: createdAt
                deviceList(cursor: $cursor) {
                    pageInfo {
                        endCursor
                    }
                    devices {
                        id: deviceShortId
                        name: alias
                        created_at: createdAt
                        device_type: deviceType {
                            id
                        }
                        disabled
                        disabled_at: disabledAt
                        updated_at: updatedAt
                    }
                }
            }
        }
    '''

    patient_id = Patient.normalize_id(patient_id)
    patient_attrs = {}

    next_cursor = None
    device_set = DeviceSet()

    # Use cursor to page through all patient devices
    while True:
        result = client.execute(
            statement=query,
            patient_id=patient_id,
            cursor=next_cursor
        )

        patient_attrs = result["patient"]

        # Add the patient's devices to device_set
        device_list = patient_attrs.get("deviceList", {})
        for device_attrs in device_list.get("devices", []):
            device_type = device_attrs['device_type']
            device_attrs['device_type_id'] = device_type['id']
            del device_attrs['device_type']

            device = Device(
                patient_id=patient_id,
                **device_attrs
            )
            device_set.add(device)

        # next_cursor is None when there are no more devices for this patient
        next_cursor = device_list.get("pageInfo", {}).get("endCursor")
        if not next_cursor:
            break

    del patient_attrs["deviceList"]
    return Patient(devices=device_set, **patient_attrs)


def get_all_patients(client: Optional[GraphClient] = None) -> PatientSet:
    """
    Get a set of all patients the user has access to.

    Args:
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    patients_query = '''
        query getPatientList($patient_cursor: Cursor, $device_cursor: Cursor) {
            org {
                patientAccessList(cursor: $patient_cursor) {
                    pageInfo {
                        endCursor
                    }
                    patientAccess {
                        patient {
                            id
                            name: codeName
                            created_at: createdAt
                            deviceList(cursor: $device_cursor) {
                                pageInfo {
                                    endCursor
                                }
                                devices {
                                    id: deviceShortId
                                    name: alias
                                    created_at: createdAt
                                    device_type: deviceType {
                                        id
                                    }
                                    disabled
                                    disabled_at: disabledAt
                                    updated_at: updatedAt
                                }
                            }
                        }
                    }
                }
            }
        }
    '''

    patient_cursor = None
    patient_set = PatientSet()

    # Use cursor to page through all patients
    while True:
        result = client.execute(
            statement=patients_query,
            patient_cursor=patient_cursor,
            device_cursor=None,
        )

        # Loop over all accessible patients
        curr_pal = result.get("org", {}).get("patientAccessList", {})
        for patient_info in curr_pal.get("patientAccess", []):
            patient_attrs = patient_info["patient"]
            patient_id = Patient.normalize_id(patient_attrs["id"])

            # Add the patient's devices to device_set
            device_set = DeviceSet()
            device_list = patient_attrs.get("deviceList", {})
            for device_attrs in device_list.get("devices", []):
                device_type = device_attrs['device_type']
                device_attrs['device_type_id'] = device_type['id']
                del device_attrs['device_type']

                device = Device(
                    patient_id=patient_id,
                    **device_attrs
                )
                device_set.add(device)

            device_cursor = device_list.get("pageInfo", {}).get("endCursor")
            if not device_cursor:
                # If there are no more devices for this patient, device_set is
                # complete and the patient can be added to the patient_set.
                del patient_attrs["deviceList"]
                patient = Patient(devices=device_set, **patient_attrs)
                patient_set.add(patient)
            else:
                # If there are more devices for this patient, then query for
                # all their patient data at once.
                patient = get_patient(patient_id, client=client)
                patient_set.add(patient)

        # patient_cursor is None when there are no more pages of patients.
        patient_cursor = curr_pal.get("pageInfo", {}).get("endCursor")
        if not patient_cursor:
            break

    return patient_set


def get_device(
    patient: Union[Patient, str],
    device_id: str,
    client: Optional[GraphClient] = None,
) -> Device:
    """
    Get a patient's device, by the device ID.

    Note that if a Patient object is provided, this function serves
    as a wrapper around Patient.device(). If a patient ID is provided, metadata
    is fetched from the API.

    Args:
        patient: a patient ID or Patient object
        device_id: Device ID
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    if type(patient) is str:
        patient = get_patient(patient_id=patient, client=client)

    return patient.device(device_id=device_id)


def get_patient_devices(
    patient: Union[Patient, str],
    client: Optional[GraphClient] = None,
) -> DeviceSet:
    """
    Get all devices for a patient.

    Note that if a Patient object is provided, this function serves
    as a wrapper around Patient.devices. If a patient ID is provided,
    metadata is fetched from the API.

    Args:
        patient: a patient ID or Patient object
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    if type(patient) is str:
        patient = get_patient(patient_id=patient, client=client)

    return patient.devices


def get_all_devices(
    patients: Union[PatientSet, List[str]] = None,
    client: Optional[GraphClient] = None,
) -> DeviceSet:
    """
    Get a set of all devices belonging to a set of patients. If a specific
    set is not specified, returns all devices belonging to all patients
    in the user's active organization.

    Args:
        patients: a list of patient IDs or a PatientSet
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    if not patients:
        return get_all_patients(client=client).devices

    if type(patients) is list:
        all_devices = DeviceSet()

        for patient_id in patients:
            patient = get_patient(patient_id=patient_id, client=client)
            for device in patient.devices:
                all_devices.add(device)

        return all_devices

    return patients.devices



"""
Fetch metadata about projects.

"""

from typing import Iterable, Optional, Type
import pandas as pd

from .client import GraphClient, global_graph_client
from .common import ItemBase, ItemSet


class Metric(ItemBase):
    """
    A measurement related to processed patient data. Each metric is related
    to a single processed data type, project, and time interval. These metrics
    are calculated periodically and updated.

    """
    def __init__(
        self,
        id: str,
        type: str,
        data_type: str,
        time_interval: str,
        updated_at: float,
        created_at: float,
        value: Optional[float] = None,
    ):
        """
        Initialize with data.

        Args:
            id: Unique identifier of the patient
            type: Type of measurement/metric. Possible types include:
                TOTAL_HOURS, LATEST_DATA, LAST_UPLOAD
            data_type: Processed stream data type that is being measured.
                Possible data types include:
                APPLEWATCH_SYMPTOM, APPLEWATCH_TREMOR,
                APPLEWATCH_DYSKINESIA, APPLEWATCH_HEART_RATE,
                PERCEPT_TREND_LOG_LFP
            value: Value of the metric (float)
            time_interval: Period over which the metric was calculated.
                Possible time intervals include:
                FOURTEEN_DAYS, NINETY_DAYS, PROJECT_ALL
            created_at: Time patient was added to the cohort (unix timestamp)
            updated_at: Time cohort patient was last updated (unix timestamp)

        """
        self.type = type
        self.data_type = data_type
        self.time_interval = time_interval
        self.created_at = created_at
        self.updated_at = updated_at
        self.value = value

        super().__init__(
            id=id,
            type=type,
            data_type=data_type,
            value=value,
            time_interval=time_interval,
            created_at=created_at,
            updated_at=updated_at,
        )


class MetricSet(ItemSet):
    """
    A collection of Metrics.

    """
    def __init__(self, items: Iterable[Metric] = ()):
        """
        Initialize with Metrics.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Metric


class ProjectPatientMetadata(ItemBase):
    """
    A patient who has been placed within a project and is now
    associated with the project.

    """
    def __init__(
        self,
        id: str,
        updated_at: float,
        created_at: float,
        created_by: str,
        updated_by: str,
        metrics: MetricSet,
        **attributes
    ):
        """
        Initialize with data.

        Args:
            id: Patient ID of the patient in the cohort
            created_at: Time patient was added to the project (unix timestamp)
            created_by: Display name of who added the patient to the project
            updated_at: Time project patient was last updated (unix timestamp)
            updated_by: Display name of who updated the project patient record
            metrics: Project data metrics related to the patient's data
            **attributes: Other attributes associated with the project

        """
        self.created_at = created_at
        self.updated_at = updated_at
        self.created_by = created_by
        self.updated_by = updated_by
        self.metrics = metrics

        super().__init__(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            created_by=created_by,
            updated_by=updated_by,
            metrics=metrics,
            **attributes,
        )

    def to_dict(self) -> dict:
        """
        Dictionary representation of the Project Patient attributes.

        """
        attrs = self._attributes.copy()
        attrs["metrics"] = self.metrics.to_list()
        return attrs

    def get_patient_metadata_dataframe(self) -> pd.DataFrame:
        """
        Returns a new dataframe displaying the patient's metadata.

        """
        column_name_map = {
            "type": "metric_type",
            "data_type": "metric_data_type",
            "value": "metric_value",
            "time_interval": "metric_time_interval",
            "created_at": "metric_created_at",
            "updated_at": "metric_updated_at",
            "id": "metric_id",
        }
        metadata_obj = self.to_dict()
        metadata_df = pd.DataFrame(metadata_obj).rename(
            columns={"code_name": "project_patient_code_name"}
        )

        return pd.concat(
            [
                metadata_df.drop('metrics', axis=1),
                pd.DataFrame(metadata_df['metrics'].tolist()).rename(
                    columns=column_name_map
                )
            ],
            axis=1
        )


class CohortPatientMetadata(ProjectPatientMetadata):
    """
    Cohort related information about a patient contained in a cohort.

    """
    def __init__(
        self,
        id: str,
        updated_at: float,
        created_at: float,
        created_by: str,
        updated_by: str,
        metrics: MetricSet,
        **attributes
    ):
        """
        Initialize with data.

        Args:
            id: Patient ID of the patient in the cohort
            created_at: Time patient was added to the cohort (unix timestamp)
            created_by: Display name of who added the patient to the cohort
            updated_at: Time cohort patient was last updated (unix timestamp)
            updated_by: Display name of who updated the cohort patient record
            metrics: Project data metrics related to the patient's data
            **attributes: Other attributes associated with the cohort

        """
        super().__init__(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            created_by=created_by,
            updated_by=updated_by,
            metrics=metrics,
            **attributes,
        )


class CohortPatientMetadataSet(ItemSet):
    """
    A collection of CohortPatientMetadata.

    """

    def __init__(self, items: Iterable[CohortPatientMetadata] = ()):
        """
        Initialize with CohortPatientMetadata.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return CohortPatientMetadata


class ProjectPatientMetadataSet(ItemSet):
    """
    A collection of ProjectPatientMetadata.

    """

    def __init__(self, items: Iterable[ProjectPatientMetadata] = ()):
        """
        Initialize with ProjectPatientMetadata.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return ProjectPatientMetadata


class Cohort(ItemBase):
    """
    A generic sub-container for a group of patients within a project.

    A cohort is a generic container within a project for a group of patients
    with workflow-related metadata: description, milestones dates,
    metrics, etc.

    A single project can have multiple cohorts.

    """
    def __init__(
        self,
        id: str,
        title: str,
        updated_at: float,
        created_at: float,
        created_by: str,
        updated_by: str,
        **attributes
    ):
        """
        Initialize with data.

        Args:
            id: ID of the project
            title: Human-readable name
            created_at: Time the cohort was created (unix timestamp)
            created_by: Display name of who created the cohort
            updated_at: Time the cohort was last updated (unix timestamp)
            updated_by: Display name of who updated the cohort
            **attributes: Other attributes associated with the cohort

        """
        self.title = title
        self.created_at = created_at
        self.updated_at = updated_at
        self.created_by = created_by
        self.updated_by = updated_by

        super().__init__(
            id=id,
            title=title,
            created_at=created_at,
            updated_at=updated_at,
            created_by=created_by,
            updated_by=updated_by,
            **attributes,
        )


class CohortSet(ItemSet):
    """
    A collection of Cohorts.

    """

    def __init__(self, items: Iterable[Cohort] = ()):
        """
        Initialize with Cohorts.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Cohort


class Project(ItemBase):
    """
    Metadata for a project.

    A project is a generic container for a group of patients
    with workflow-related metadata: status, project type, description,
    milestones dates, etc. Data availability QC metrics are computed
    on a regular basis for all patients in a project.

    """
    def __init__(
        self,
        id: str,
        title: str,
        status: str,
        type: str,
        started_at: float,
        updated_at: float,
        created_at: float,
        created_by: str,
        updated_by: str,
        cohorts: CohortSet(),
        **attributes
    ):
        """
        Initialize with data.

        Args:
            id: ID of the project
            title: Human-readable name
            status: Status of the project
            type: Type of project. Possible types include:
                EXPLORATORY, CLINICAL_TRIAL,RETROSPECTIVE_STUDY,
                PROSPECTIVE_STUDY, SANDBOX
            started_at: Time the project started (unix timestamp)
            created_at: Time the project was created (unix timestamp)
            created_by: Display name of who created the project
            updated_at: Time the project was last updated (unix timestamp)
            updated_by: Display name of who updated the project
            cohorts: Sub-containers of patients in a project
            **attributes: Other attributes associated with the project

        """
        self.title = title
        self.created_at = created_at
        self.started_at = started_at
        self.updated_at = updated_at
        self.created_by = created_by
        self.updated_by = updated_by
        self.status = status
        self.type = type
        self.cohorts = cohorts

        super().__init__(
            id=id,
            title=title,
            created_at=created_at,
            started_at=started_at,
            updated_at=updated_at,
            created_by=created_by,
            updated_by=updated_by,
            status=status,
            type=type,
            cohorts=cohorts,
            **attributes,
        )

    def to_dict(self) -> dict:
        """
        Dictionary representation of the Project attributes.

        """
        attrs = self._attributes.copy()
        attrs["cohorts"] = self.cohorts.to_list()

        return attrs


class ProjectSet(ItemSet):
    """
    A collection of Projects.

    """

    def __init__(self, items: Iterable[Project] = ()):
        """
        Initialize with Projects.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Project


def get_project(
    project_id: str, client: Optional[GraphClient] = None
) -> Project:
    """
    Get the project with the specified ID.

    Args:
        project_id: Project ID
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    query = """
        query getProject($id: ID) {
            project (id: $id) {
                id,
                title,
                status,
                description,
                type,
                cohortList {
                    cohorts {
                        id,
                        title,
                        description,
                        created_at: createdAt,
                        updated_at: updatedAt,
                        created_by: createdBy,
                        updated_by: updatedBy,
                    }
                },
                created_at: createdAt,
                updated_at: updatedAt,
                started_at: startedAt,
                created_by: createdBy,
                updated_by: updatedBy
            }
        }
    """

    result = client.execute(statement=query, id=project_id)

    project_attrs = result["project"]

    # Move cohorts to attribute level
    cohorts = project_attrs.get("cohortList", {}).get("cohorts", [])
    cohort_set = CohortSet()

    for cohort in cohorts:
        cohort_obj = Cohort(**cohort)
        cohort_set.add(cohort_obj)

    project_attrs["cohorts"] = cohort_set
    del project_attrs["cohortList"]

    return Project(**project_attrs)


def get_projects(client: Optional[GraphClient] = None) -> ProjectSet:
    """
    Get all the projects that the current user has access to.

    Args:
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    query = """
        query($cursor: DateTimeUUIDCursor) {
            org {
                id
                projectList(cursor: $cursor) {
                    projects {
                        id,
                        title,
                        status,
                        description,
                        type,
                        cohortList {
                            cohorts {
                                id,
                                title,
                                description,
                                created_at: createdAt,
                                updated_at: updatedAt,
                                created_by: createdBy,
                                updated_by: updatedBy,
                            }
                        },
                        created_at: createdAt,
                        updated_at: updatedAt,
                        started_at: startedAt,
                        created_by: createdBy,
                        updated_by: updatedBy
                    }
                    pageInfo {
                        endCursor
                    }
                }
            }
        }
    """

    next_cursor = None
    project_set = ProjectSet()

    # Use cursor to page through all projects
    while True:
        result = client.execute(statement=query, cursor=next_cursor)
        project_list = result.get("org", {}).get("projectList", {})

        for project in project_list.get("projects", []):
            cohorts = project.get("cohortList", {}).get("cohorts", [])
            cohort_set = CohortSet()

            for cohort in cohorts:
                cohort_obj = Cohort(**cohort)
                cohort_set.add(cohort_obj)

            project["cohorts"] = cohort_set
            del project["cohortList"]

            project = Project(**project)
            project_set.add(project)

        # endCursor is None when there are no more pages of data
        next_cursor = project_list.get("pageInfo", {}).get("endCursor")
        if not next_cursor:
            break

    return project_set


def _parse_patient_metrics(
    metrics: dict
) -> MetricSet:
    """
    Helper function that intakes a metric dictionary and outputs a metric set.

    """
    metric_set = MetricSet()

    for metric in metrics:
        metric = Metric(**metric)
        metric_set.add(metric)

    return metric_set


def get_project_patients(
    project_id: str,
    client: Optional[GraphClient] = None,
) -> ProjectPatientMetadataSet:
    """
    Get all patients in a project and their associated project data metrics.

    Args:
        project_id: ID of the project
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    project_patients_query = """
        query($id: ID, $cursorInput: CursorInput) {
            project(id: $id) {
                projectPatientList(
                    cursorInput: $cursorInput
                ){
                    projectPatients{
                        patient {
                            id
                        },
                        metricList {
                            metrics {
                                id,
                                updated_at: updatedAt,
                                created_at: createdAt,
                                type,
                                data_type: dataType,
                                value,
                                time_interval: timeInterval
                            }
                        },
                        code_name: codeName,
                        created_at: createdAt,
                        updated_at: updatedAt,
                        created_by: createdBy,
                        updated_by: updatedBy,
                    },
                    pageInfo{
                        codeNameEndCursor
                    }
                },
            }
        }
    """

    cursor_input = None
    project_patient_set = ProjectPatientMetadataSet()

    # Use cursor to page through all patients
    while True:
        result = client.execute(
            statement=project_patients_query,
            id=project_id,
            cursor_input=cursor_input,
        )

        # Loop over all project patients
        patient_list = result.get("project", {}).get("projectPatientList")

        for patient_info in patient_list.get("projectPatients", []):
            patient_attrs = patient_info

            # Move patient id to top level attribute
            patient_attrs["id"] = patient_attrs.get("patient").get("id")
            del patient_attrs["patient"]

            metrics = patient_attrs.get("metricList").get("metrics")
            patient_attrs["metrics"] = _parse_patient_metrics(metrics)
            del patient_attrs["metricList"]

            patient = ProjectPatientMetadata(**patient_attrs)
            project_patient_set.add(patient)

        # cursor is None when there are no more pages of project patients.
        cursor = patient_list.get("pageInfo", {}).get("codeNameEndCursor")
        if not cursor:
            break

        cursor_input = {
            "codeNameCursor": cursor
        }

    return project_patient_set


def get_cohort_patients(
    cohort_id: str,
    client: Optional[GraphClient] = None,
) -> CohortPatientMetadataSet:
    """
    Get all patients in a cohort.

    Args:
        cohort_id: ID of the cohort
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    cohort_patients_query = """
        query($id: ID, $cursorInput: CursorInput) {
            cohort(id: $id) {
                id,
                cohortPatientList(cursorInput: $cursorInput) {
                    cohortPatients {
                        patient {
                            id
                        }
                        metricList {
                            metrics {
                                id,
                                type,
                                data_type: dataType,
                                value,
                                time_interval: timeInterval,
                                created_at: createdAt,
                                updated_at: updatedAt
                            }
                        }
                        code_name: codeName,
                        created_at: createdAt,
                        updated_at: updatedAt,
                        created_by: createdBy,
                        updated_by: updatedBy,
                    }
                    pageInfo {
                        codeNameEndCursor
                    }
                }
            }
        }
    """
    cursor_input = None
    cohort_patient_set = CohortPatientMetadataSet()

    while True:
        result = client.execute(
            statement=cohort_patients_query,
            id=cohort_id,
            cursor_input=cursor_input,
        )

        cohort_patient_list = result.get("cohort").get("cohortPatientList", {})
        cohort_patients = cohort_patient_list.get("cohortPatients", [])

        for patient_info in cohort_patients:
            patient_attrs = patient_info

            # Move patient id to top level attribute
            patient_attrs["id"] = patient_attrs.get("patient").get("id")
            del patient_attrs["patient"]

            metrics = patient_attrs.get("metricList", {}).get("metrics", [])
            patient_attrs["metrics"] = _parse_patient_metrics(metrics)
            del patient_attrs["metricList"]

            patient = CohortPatientMetadata(**patient_attrs)
            cohort_patient_set.add(patient)

        cursor = cohort_patient_list.get("pageInfo", {}).get(
            "codeNameEndCursor"
        )

        if not cursor:
            break

        cursor_input = {
            "codeNameCursor": cursor
        }

    return cohort_patient_set



"""
Tests for fetching org metadata.

"""
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient
from runeq.resources.org import (
    get_org,
    get_orgs,
    Org,
    set_active_org
)


class TestOrg(TestCase):
    """
    Unit tests for the Org class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )

    def test_attributes(self):
        """
        Test attributes for an initialized Org.

        """
        test_org = Org(
            id="org1-id",
            created_at=1629300943.9179766,
            name="org1",
        )

        self.assertEqual("org1-id", test_org.id)
        self.assertEqual(1629300943.9179766, test_org.created_at)
        self.assertEqual("org1", test_org.name)

    def test_repr(self):
        """
        Test __repr__

        """
        test_org = Org(
            id="org1-id",
            created_at=1629300943.9179766,
            name="Org 1",
        )
        self.assertEqual(
            'Org(id="org1-id", name="Org 1")',
            repr(test_org)
        )

    def test_normalize_id(self):
        """
        Test strip id of resource prefix and suffix if they exist.

        """
        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485,org")
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485,org")
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485")
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Org.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485")
        )

    def test_denormalize_id(self):
        """
        Test add id resource prefix and suffix if they don't exist.

        """
        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485,org")
        )

        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("d4b1c627bd464fe0a5ed940cc8e8e485,org")
        )

        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("org-d4b1c627bd464fe0a5ed940cc8e8e485")
        )

        self.assertEqual(
            "org-d4b1c627bd464fe0a5ed940cc8e8e485,org",
            Org.denormalize_id("d4b1c627bd464fe0a5ed940cc8e8e485")
        )

    def test_get_org(self):
        """
        Test get org for specified org_id.

        """
        self.mock_client.execute = mock.Mock(return_value={
            "user": {
                "membershipList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "memberships": [
                        {
                            "org": {
                                "id": "org1-id",
                                "created_at": 1571267538.391721,
                                "name": "org1"
                            }
                        },
                        {
                            "org": {
                                "id": "org2-id",
                                "created_at": 1630515986.9949625,
                                "name": "org2"
                            }
                        }
                    ]
                }
            }
        })

        org = get_org("org1-id", client=self.mock_client)
        self.assertEqual(
            {
                "id": "org1-id",
                "created_at": 1571267538.391721,
                "name": "org1"
            },
            org.to_dict(),
        )

        org = get_org("org2-id", client=self.mock_client)
        self.assertEqual(
            {
                "id": "org2-id",
                "created_at": 1630515986.9949625,
                "name": "org2"
            },
            org.to_dict(),
        )

    def test_get_orgs_basic(self):
        """
        Test get orgs for the initialized user.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "user": {
                    "membershipList": {
                        "pageInfo": {
                            "endCursor": None
                        },
                        "memberships": [
                            {
                                "org": {
                                    "id": "org1-id",
                                    "created_at": 1571267538.391721,
                                    "name": "org1"
                                }
                            },
                            {
                                "org": {
                                    "id": "org2-id",
                                    "created_at": 1630515986.9949625,
                                    "name": "org2"
                                }
                            },
                            {
                                "org": {
                                    "id": "org3-id",
                                    "created_at": 1649888079.066764,
                                    "name": "org3"
                                }
                            }
                        ]
                    }
                }
            }
        ]

        orgs = get_orgs(client=self.mock_client)

        self.assertEqual(
            [
                {
                    'created_at': 1571267538.391721,
                    'name': 'org1',
                    'id': 'org1-id'
                },
                {
                    'created_at': 1630515986.9949625,
                    'name': 'org2',
                    'id': 'org2-id'
                },
                {
                    'created_at': 1649888079.066764,
                    'name': 'org3',
                    'id': 'org3-id'
                }
            ],
            orgs.to_list(),
        )

    def test_get_orgs_pagination(self):
        """
        Test get orgs for the initialized user, where the user has to
        page through orgs.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "user": {
                    "membershipList": {
                        "pageInfo": {
                            "endCursor": "test_check_next"
                        },
                        "memberships": [
                            {
                                "org": {
                                    "id": "org1-id",
                                    "created_at": 1571267538.391721,
                                    "name": "org1"
                                }
                            }
                        ]
                    }
                }
            },
            {
                "user": {
                    "membershipList": {
                        "pageInfo": {
                            "endCursor": None
                        },
                        "memberships": [
                            {
                                "org": {
                                    "id": "org2-id",
                                    "created_at": 1630515986.9949625,
                                    "name": "org2"
                                }
                            }
                        ]
                    }
                }
            }
        ]

        orgs = get_orgs(client=self.mock_client)

        self.assertEqual(
            [
                {
                    'created_at': 1571267538.391721,
                    'name': 'org1',
                    'id': 'org1-id'
                },
                {
                    'created_at': 1630515986.9949625,
                    'name': 'org2',
                    'id': 'org2-id'
                }
            ],
            orgs.to_list(),
        )

    def test_set_active_org(self):
        """
        Test setting the user's active organization

        """
        org_id = "org1-id"
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "updateDefaultMembership": {
                    "user": {
                        "defaultMembership": {
                            "org": {
                                "id": org_id,
                                "name": "org1",
                                "created_at": 1571267538.391721,
                            }
                        }
                    }
                }
            }
        ]

        new_org = set_active_org(org_id, self.mock_client)
        self.mock_client.execute.assert_called_once()
        self.assertEqual(
            new_org.to_dict(),
            {
                'created_at': 1571267538.391721,
                'name': 'org1',
                'id': 'org1-id'
            }
        )



from setuptools import find_packages, setup

package_name = 'runeq'

with open('README.md') as r:
    long_description = r.read()

with open('requirements/common.txt', 'r') as r:
    install_requires = r.read().splitlines()

setup(
    name=package_name,
    version='0.12.0',
    author='Rune Labs',
    maintainer_email='support@runelabs.io',
    description='Query data from Rune Labs APIs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='rune labs api query data',
    url='https://github.com/rune-labs/runeq-python',
    packages=find_packages(),
    install_requires=install_requires,
    test_suite='tests',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    entry_points={
        'console_scripts': [
            'runeq=cli:cli',
        ],
    },
)



"""
Configuration for accessing Rune APIs.

"""
import os
import yaml


AUTH_METHOD_CLIENT_KEYS = 'client_keys'
"""
Auth method to use a Client Key pair for authentication.

"""

AUTH_METHOD_JWT = 'jwt'
"""
Auth method to use a JWT for authentication.

"""

AUTH_METHOD_ACCESS_TOKEN = 'access_token'
"""
Auth method to use a user access token for authentication.

"""

DEFAULT_CONFIG_YAML = '~/.rune/config'
"""
Default path for the config file (yaml formatted)

"""


class Config:
    """
    Holds configuration (e.g. auth credentials, URLs, etc)

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize configuration options.

        Args:
            *args: Accepts at most 1; a filename. If provided, values will be
                loaded from the file, using `load_yaml()`. It is invalid to
                provide both a filename **and** keyword arguments.
            **kwargs: Passed to set_values().

        Examples:
            There are three valid ways to create a config:

            >>> cfg = Config()
            # Load from default file location (~/.rune/config)

            >>> cfg = Config('./example_config.yaml')
            # Load from a specified YAML file

            >>> cfg = Config(access_token_id='foo', access_token_secret='bar')
            # Set values using keyword arguments. This can be used with any
            # valid combination of config options; the example above sets a
            # user access token.

        """
        self.graph_url = 'https://graph.runelabs.io'
        self.stream_url = 'https://stream.runelabs.io'
        self.auth_method = None
        self._access_token_id = None
        self._access_token_secret = None
        self._client_access_key = None
        self._client_key_id = None
        self._jwt = None

        if args and kwargs:
            raise TypeError(
                '__init__() cannot accept both positional arguments and '
                'keyword arguments'
            )
        elif args:
            if len(args) > 1:
                raise TypeError(
                    '__init__() takes at most 1 positional argument but '
                    f'{len(args)} were given')

            self.load_yaml(args[0])
        elif not kwargs:
            self.load_yaml()
        else:
            self.set_values(**kwargs)

    def load_yaml(self, filename=DEFAULT_CONFIG_YAML):
        """
        Set values from a YAML file. Keys from the file are passed directly to
        set_values(), as kwargs.

        Args:
            filename: File path for a YAML-formatted config file

        """
        with open(os.path.expanduser(filename)) as f:
            params = yaml.safe_load(f)
        return self.set_values(**params)

    def set_values(
        self,
        auth_method=None,
        access_token_id=None,
        access_token_secret=None,
        client_key_id=None,
        client_access_key=None,
        jwt=None,
        stream_url=None,
        graph_url=None,
        **kwargs,
    ):
        """
        Set configuration values.

        Args:
            auth_method: What type of authentication credentials to use. Must
                be one of 'access_token', 'client_keys', or 'jwt'. If not set,
                the auth method is inferred based on which credentials are
                specified (as long as it's unambiguous).
            access_token_id: User access token ID
            access_token_secret: User access token secret
            client_key_id: Client key ID
            client_access_key: Client access key
            jwt: JWT
            stream_url: base URL to use for the stream API
            graph_url: base URL to use for the graph API
            **kwargs: Arbitrary values may be provided, but they are ignored.

        """
        if stream_url is not None:
            self.stream_url = stream_url

        if graph_url is not None:
            self.graph_url = graph_url

        if auth_method is None:
            num_auth_methods_set = sum([
                bool(access_token_id or access_token_secret),
                bool(client_access_key or client_key_id),
                bool(jwt)
            ])

            if num_auth_methods_set > 1:
                raise ValueError(
                    'Cannot infer auth method: multiple credentials were '
                    'provided. Specify auth_method kwarg to disambiguate.'
                )
            elif access_token_id and access_token_secret:
                auth_method = AUTH_METHOD_ACCESS_TOKEN
            elif client_key_id and client_access_key:
                auth_method = AUTH_METHOD_CLIENT_KEYS
            elif jwt:
                auth_method = AUTH_METHOD_JWT
            else:
                raise ValueError(
                    'Cannot infer auth method: a complete set of credentials '
                    'was not provided.'
                )

        self.auth_method = auth_method

        if access_token_id is not None:
            self._access_token_id = access_token_id

        if access_token_secret is not None:
            self._access_token_secret = access_token_secret

        if client_key_id is not None:
            self._client_key_id = client_key_id

        if client_access_key is not None:
            self._client_access_key = client_access_key

        if jwt is not None:
            self._jwt = jwt

        # access auth headers to ensure they're valid
        _ = self.auth_headers

    @property
    def client_auth_headers(self):
        """
        Authentication headers for HTTP requests, using client key pair.

        """
        if not self._client_access_key:
            raise ValueError('Client access key is not set')
        if not self._client_key_id:
            raise ValueError('Client key id is not set')

        return {
            'X-Rune-Client-Key-ID': self._client_key_id,
            'X-Rune-Client-Access-Key': self._client_access_key,
        }

    @property
    def jwt_auth_headers(self):
        """
        Authentication headers for HTTP requests, using a JWT.

        """
        if not self._jwt:
            raise ValueError('JWT is not set')

        return {
            'X-Rune-User-Access-Token': self._jwt,
        }

    @property
    def access_token_auth_headers(self):
        """
        Authentication headers for HTTP requests, using a User access token.

        """
        if not self._access_token_id:
            raise ValueError('Access token id is not set')

        if not self._access_token_secret:
            raise ValueError('Access token secret is not set')

        return {
            'X-Rune-User-Access-Token-Id': self._access_token_id,
            'X-Rune-User-Access-Token-Secret': self._access_token_secret
        }

    @property
    def auth_headers(self):
        """
        Authentication headers for HTTP requests to Rune APIs.

        """
        if self.auth_method == AUTH_METHOD_ACCESS_TOKEN:
            return self.access_token_auth_headers
        elif self.auth_method == AUTH_METHOD_CLIENT_KEYS:
            return self.client_auth_headers
        elif self.auth_method == AUTH_METHOD_JWT:
            return self.jwt_auth_headers
        else:
            raise ValueError(
                f'Invalid auth_method "{self.auth_method}": expected one of '
                f'({AUTH_METHOD_CLIENT_KEYS}, {AUTH_METHOD_JWT})'
            )



#!/usr/bin/env python3
"""
runeq command line interface

"""
import os
import yaml

import click

from runeq.config import DEFAULT_CONFIG_YAML


def _write_config(values):
    """
    Write values to the configuration file. Overwrites any existing content.

    """
    filepath = os.path.expanduser(DEFAULT_CONFIG_YAML)

    # Create the directory if it doesn't exist
    dirpath = os.path.dirname(filepath)
    if not os.path.isdir(dirpath):
        os.mkdir(dirpath)

    with open(filepath, 'w') as f:
        yaml.dump(values, f)


def _get_config() -> dict:
    """
    Get or create the config file

    """
    filepath = os.path.expanduser(DEFAULT_CONFIG_YAML)
    if not os.path.exists(filepath):
        return {}

    try:
        with open(filepath) as f:
            values = yaml.safe_load(f)

        valid = isinstance(values, dict) or not values
    except yaml.YAMLError:
        valid = False

    if valid:
        return values or {}

    # If the contents of the config file are not a dictionary, ask the user
    # if they want to overwrite the contents.
    prompt = (
        'Detected invalid contents in the config file ({})\nWould you like to '
        'reset the contents?'.format(filepath)
    )
    if not click.confirm(click.style(prompt, fg='red')):
        raise click.Abort()

    _write_config({})
    return {}


@click.group()
def cli():
    """
    runeq command line interface

    """


@cli.group()
def configure():
    """
    Manage runeq configuration (stored in a YAML file).

    """


@configure.command()
@click.option('--access-token-id', required=True,
              prompt='Enter an access token ID',
              help='Rune access token ID')
@click.option('--access-token-secret', required=True,
              prompt='Enter an access token secret',
              help='Rune access token secret')
def setup(access_token_id, access_token_secret):
    """
    Set up a runeq configuration file. This will overwrite any existing values.

    """
    if _get_config():
        prompt = (
            'Found existing config values. Would you like to overwrite them?'
        )
        if not click.confirm(click.style(prompt, fg='red')):
            raise click.Abort()

    _write_config({
        'access_token_id': access_token_id,
        'access_token_secret': access_token_secret,
    })
    click.echo(click.style(
        f'Success! Created a configuration file: {DEFAULT_CONFIG_YAML}',
        fg='green'
    ))


@configure.command()
@click.option('--key', '-k', default='', multiple=True, required=False,
              help='1+ keys to fetch')
def get(key):
    """
    Print one or more values from the current runeq configuration. If no keys
    are specified, prints all existing values.

    """
    config = _get_config()

    if not key:
        for k in sorted(config):
            print('{}: {}'.format(k, config[k]))
    else:
        for k in key:
            print('{}: {}'.format(k, config.get(k, '')))


@configure.command()
@click.option('--value', '-v', multiple=True, required=True,
              help='1+ values to set in the config file. Each option should'
                   'be provided as [key]=[value]')
def set(value):
    """
    Set a value in the runeq configuration file.

    """
    config = _get_config()

    for item in value:
        elements = item.split('=')
        if len(elements) != 2:
            raise click.ClickException(
                'Each option must be provided as [key]=[value]'
            )

        k, v = elements
        config[k.strip()] = v.strip()

    _write_config(config)


if __name__ == '__main__':
    cli()



"""
Fetch metadata about organizations. A research lab or clinical site is
typically represented as an organization.

"""
from typing import Iterable, Optional, Type, Union

from .client import GraphClient, global_graph_client
from .common import ItemBase, ItemSet


class Org(ItemBase):
    """
    Metadata for an organization.

    """

    def __init__(
        self,
        id: str,
        name: str,
        created_at: float,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: ID of the organization
            name: Human-readable name
            created_at: When the organization was created (unix timestamp)
            **attributes: Other attributes associated with the organization

        """
        norm_id = Org.normalize_id(id)
        self.name = name
        self.created_at = created_at

        super().__init__(
            id=norm_id,
            name=name,
            created_at=created_at,
            **attributes,
        )

    @staticmethod
    def normalize_id(org_id: str) -> str:
        """
        Strip resource prefix and suffix from an org ID (if they exist).

        Args:
            org_id: Organization ID

        """
        if org_id.startswith("org-"):
            org_id = org_id[4:]

        if org_id.endswith(",org"):
            org_id = org_id[:-4]

        return org_id

    @staticmethod
    def denormalize_id(org_id: str) -> str:
        """
        Add resource prefix and suffix to an org ID (if they don't exist).

        This constructs the form of the ID that is used for requests to the
        GraphQL API.

        Args:
            org_id: Organization ID

        """
        if not org_id.startswith("org-"):
            org_id = f'org-{org_id}'

        if not org_id.endswith(",org"):
            org_id = f'{org_id},org'

        return org_id


class OrgSet(ItemSet):
    """
    A collection of Organizations.

    """

    def __init__(self, items: Iterable[Org] = ()):
        """
        Initialize with Orgs.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Org


def _iter_all_orgs(client: Optional[GraphClient] = None):
    """
    Fetch all orgs that the current user is a member of, and yield each
    one (while paging through the response)

    """
    client = client or global_graph_client()
    query = '''
        query getOrgMemberships($cursor: Cursor) {
            user {
                membershipList(cursor: $cursor) {
                    pageInfo {
                        endCursor
                    }
                    memberships {
                        org {
                            id
                            created_at: created
                            name: displayName
                        }
                    }
                }
            }
        }
    '''

    next_cursor = None

    # Use cursor to page through all orgs that the user is a member of. Yield
    # each one as an Org
    while True:
        result = client.execute(
            statement=query,
            cursor=next_cursor
        )

        membership_list = result.get("user", {}).get("membershipList", {})

        for membership in membership_list.get("memberships", []):
            org_attrs = membership["org"]
            org = Org(
                id=org_attrs["id"],
                created_at=org_attrs.get("created_at"),
                name=org_attrs.get("name"),
            )
            yield org

        # endCursor is None when there are no more pages of data
        next_cursor = membership_list.get("pageInfo", {}).get("endCursor")
        if not next_cursor:
            break


def get_org(org_id: str, client: Optional[GraphClient] = None) -> Org:
    """
    Get the org with the specified ID.

    Args:
        org_id: Organization ID
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    org_id = Org.normalize_id(org_id)
    # The getOrg query only works for the user's currently active organization.
    # To make this work for any org that the user is a member of, use
    # getOrgMemberships and try to find a matching org ID. It's slow, but it
    # works for now.
    for org in _iter_all_orgs(client):
        if org_id == org.id:
            return org

    raise ValueError(f"Org not found with id: {org_id}")


def get_orgs(client: Optional[GraphClient] = None) -> OrgSet:
    """
    Get all the organizations that the current user is a member of.

    Args:
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """

    org_set = OrgSet()
    for org in _iter_all_orgs(client):
        org_set.add(org)

    return org_set


def set_active_org(
    org: Union[str, Org],
    client: Optional[GraphClient] = None
) -> Org:
    """
    Set the active organization for the current user.

    Args:
        org: an org ID or Org
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    Returns:
        The active organization

    """
    client = client or global_graph_client()
    org_id = org.id if isinstance(org, Org) else org

    statement = '''
    mutation updateDefaultMembership($input: UpdateDefaultMembershipInput!) {
        updateDefaultMembership(input: $input) {
            user {
                defaultMembership {
                    org {
                        id
                        created_at: created
                        name: displayName
                    }
                }
            }
        }
    }
    '''

    result = client.execute(statement=statement, input={"orgId": org_id})

    user_attrs = result.get("updateDefaultMembership", {}).get("user", {})
    org_attrs = user_attrs.get("defaultMembership", {}).get("org")

    new_org = Org(
        id=org_attrs["id"],
        created_at=org_attrs.get("created_at"),
        name=org_attrs.get("name"),
    )

    return new_org
